using System;
using System.Collections.Generic;
using Nemerle.Collections;
using Nemerle.Compiler;
using Nemerle.Compiler.Parsetree;
using Nemerle.Utility;

using NST = Nemerle.Compiler.NamespaceTree;

namespace Oyster.Units.Macros
{
	/// Used at compile time by Units macros - stores some intermediate information
	module CompileTimeUtil
	{
		/// Units NS name
		public UnitsNamespace = "Oyster.Units";

		/// Stores list of base units (such as mass or length) registered in the system
		public mutable BaseUnits : list[string];

		/// Stores list of all registered units including aliases
		public mutable AllUnits : list[string];

		/// Stores powers list for given base unit or alias
		public mutable PowersMap : Dictionary[string, list[int]] = Dictionary();
		

		/// Generates (or gets) given unit type with given or automatically generated name,
		/// given base units powers and given normalization coefficient
		public GetUnitType(env : GlobalEnv, typeName : string, baseUnitPowers : list[int], normalizationCoef : double) : TypeInfo
		{
			// Determine name of the type to be generated
			def typeName = match (typeName)
			{
				| null => Macros.NewSymbol("TempUnitType")
				| _ => Macros.UseSiteSymbol(typeName)
			};
			
			// Try to find this type by name among existing ones
			def unitType = env.LookupType([ typeName.Id ]);
			if (unitType.HasValue)
			{
				unitType.Value
			}
			else
			{
				// Generate new type
				def abstractTypeName = Macros.UseSiteSymbol(GetAbstractUnitType(baseUnitPowers).Name);
				def unitType = DefineInCustomNamespace(<[ decl:
					[ConcreteUnitType($(normalizationCoef : double))]
					public class $(typeName : name) : $(abstractTypeName : name)
					{
						public this(value : $(abstractTypeName : name))
						{
							base(value)
						}
						
						public this(value : double)
						{
							base(value)
						}
						
						
						override public NormalizationCoef : double
						{
							get { $(normalizationCoef : double) }
						}
						
            
            static public @:>(value : $(typeName : name)) : double
            {
              value.Value
            }
            
            static public @:(value : double) : $(typeName : name)
            {
              $(typeName : name)(value)
            }


						static public @*(unit : $(typeName : name), value : double) : $(typeName : name)
						{
							$(typeName : name)(unit.Value * value)
						}

						static public @/(unit : $(typeName : name), value : double) : $(typeName : name)
						{
							$(typeName : name)(unit.Value / value)
						}

						static public @+(unit : $(typeName : name), value : double) : $(typeName : name)
						{
							$(typeName : name)(unit.Value + value)
						}

						static public @-(unit : $(typeName : name), value : double) : $(typeName : name)
						{
							$(typeName : name)(unit.Value - value)
						}


						static public @*(value : double, unit : $(typeName : name)) : $(typeName : name)
						{
							$(typeName : name)(value * unit.Value)
						}

						static public @/(value : double, unit : $(typeName : name)) : $(typeName : name)
						{
							$(typeName : name)(value / unit.Value)
						}

						static public @+(value : double, unit : $(typeName : name)) : $(typeName : name)
						{
							$(typeName : name)(value + unit.Value)
						}

						static public @-(value : double, unit : $(typeName : name)) : $(typeName : name)
						{
							$(typeName : name)(value - unit.Value)
						}


						static public @++(value : $(typeName : name)) : $(typeName : name)
						{
							$(typeName : name)(value.Value + 1)
						}

						static public @--(value : $(typeName : name)) : $(typeName : name)
						{
							$(typeName : name)(value.Value - 1)
						}

						static public @+(value : $(typeName : name)) : $(typeName : name)
						{
							value
						}

						static public @-(value : $(typeName : name)) : $(typeName : name)
						{
							$(typeName : name)(0 - value.Value)
						}
					}
				]>);

				unitType.Compile();
				unitType
			}
		}
		
		/// Generates (or gets) abstract unit type for given base unit powers
		GetAbstractUnitType(baseUnitPowers : list[int]) : TypeInfo
		{
			def env = GlobalEnv.Core;

			// Try to find given type
			def abstractTypeName = $"$UnitsNamespace.AbstractUnitType$(GenerateUnitTypeSuffix(baseUnitPowers))";
			def abstractType = env.LookupType([ abstractTypeName ]);
			if (abstractType.HasValue)
			{
				abstractType.Value
			}
			else
			{
				// Generate abstract type
				def abstractType = DefineInCustomNamespace(<[ decl:
					abstract public class $(abstractTypeName : usesite) : UnitTypeBase
					{
						protected this(value : $(abstractTypeName : usesite))
						{
							base(value)
						}
						
						protected this(value : double)
						{
							base(value)
						}
					}
				]>);
				
				abstractType.Compile();
				abstractType
			}
		}


		/// Defines given class in custom global namespace
		public DefineInCustomNamespace(decl : ClassMember.TypeDeclaration) : TypeBuilder
		{
			def builder = GlobalEnv.Core.Define(decl);

			// Register newly created type manually - otherwise it will be unbound
			NST.ExactPath(NString.Split(builder.Name, '.')).Value = NST.TypeInfoCache.Cached(builder);

			builder
		}

		/// Increases power for given unit into list
		public ModifyBaseUnitPower(baseUnitPowers : list[int], unitName : string, f : int -> int) : list[int]
		{
			// Maybe prepare list full of zeroes
			def baseUnitPowers =
				if (baseUnitPowers.Length != 0) baseUnitPowers else List.Map(BaseUnits, fun(_) { 0 });

			List.Map2(BaseUnits, baseUnitPowers, fun(n, i) { if (n == unitName) f(i) else i })
		}

		/// Generates abstract unit type suffix basing on base unit powers
		public GenerateUnitTypeSuffix(baseUnitPowers : list[int]) : string
		{
			| [] => string.Empty
			| _ => "_" + baseUnitPowers.Head.ToString() + GenerateUnitTypeSuffix(baseUnitPowers.Tail)
		}

		/// Parses abstract unit type suffix
		public ParseUnitTypeSuffix(typeName : string) : list[int]
		{
			List.Map(NString.Split(typeName.Substring(typeName.IndexOf('_') + 1), '_'), int.Parse)
		}

		/// Returns PExpr type
		public GetExprType(ctx : Typer, expr : PExpr) : TypeInfo
		{
			ctx.TypeExpr(expr).ty.FixedValue.TypeInfo
		}

		/// Returns normalization coef if type is concrete unit type
		public GetUnitTypeCoef(t : TypeInfo) : option[double]
		{
			| _ is TypeBuilder =>
				// Get attribute using Nemerle approach
				mutable value = None();
				foreach (attr in t.GetModifiers().GetCustomAttributes())
				{
					| <[ ConcreteUnitType($(coef : double)) ]> => value = Some(coef)
				};
				value
			| _ =>
				// "Usual" attribute gathering
				def attr = Attribute.GetCustomAttribute(t.SystemType, typeof(ConcreteUnitTypeAttribute));
				def attr = attr :> ConcreteUnitTypeAttribute;
				if (attr == null) None() else Some(attr.NormalizationCoef)
		}


		/// Used inside ariphmetic op macros
		public OpMacroBody(
			ctx : Typer,
			lexpr : PExpr,
			rexpr : PExpr,
			opName : string,
			opExpr : PExpr,
			someFun : TypeInfo * TypeInfo * double * double -> PExpr) : PExpr
		{
			def ltype = GetExprType(ctx, lexpr);
			def rtype = GetExprType(ctx, rexpr);
			match ((GetUnitTypeCoef(ltype), GetUnitTypeCoef(rtype)))
			{
				| (Some(lcoef), Some(rcoef)) =>
					someFun(ltype, rtype, lcoef, rcoef)
				| _ =>
					def macroPath = NST.ExactPath(["Oyster", "Units", "Macros", opName]);
					def value = macroPath.Value;
					macroPath.Value = NST.TypeInfoCache.No();
					def resexpr = ctx.TypeExpr(opExpr);
					macroPath.Value = value;
					<[ $(resexpr : typed) ]>
			}
		}

		/// Used inside * and / macros
		public MulDivMacroBody(
			ctx : Typer,
			lexpr : PExpr,
			rexpr : PExpr,
			opName : string,
			opExpr : PExpr,
			powersFun : int * int -> int,
			coefFun : double * double -> double,
			coefOpExpr : PExpr) : PExpr
		{
			OpMacroBody(
				ctx,
				lexpr,
				rexpr,
				opName,
				opExpr,
				fun (ltype, rtype, lcoef, rcoef)
				{
					def powers = List.Map2(
						ParseUnitTypeSuffix(ltype.BaseType.Name),
						ParseUnitTypeSuffix(rtype.BaseType.Name),
						powersFun);
					def concreteType = GetUnitType(ctx.Env, null, powers, 1.0);
					<[ $(concreteType.FullName : usesite)(double.op_Multiply($coefOpExpr, $(coefFun(lcoef, rcoef) : double))) ]>
				})
		}

		/// Used inside + and - macros
		public AddSubMacroBody(
			ctx : Typer,
			lexpr : PExpr,
			rexpr : PExpr,
			opName : string,
			opExpr : PExpr,
			ctorFun : PExpr -> PExpr) : PExpr
		{
			OpMacroBody(
				ctx,
				lexpr,
				rexpr,
				opName,
				opExpr,
				fun (ltype, rtype, lcoef, rcoef)
				{
					match (ltype.BaseType.Name)
					{
						| s when s == rtype.BaseType.Name =>
							def coef = rcoef / lcoef;
							def paramExpr = ctorFun(<[ double.op_Multiply($rexpr.Value, $(coef : double)) ]>);
							<[ $(ltype.FullName : usesite)($paramExpr) ]>
						| _ =>
							Message.FatalError(
								$"Expression { $lexpr [$opName] $rexpr } can't be compiled "
								 "because units are having different dimensions.")
					}
				})
		}
	}
}
