using System;
using System.Collections.Generic;
using Nemerle.Collections;
using Nemerle.Compiler;
using Nemerle.Macros;

using System.Console;

namespace Oyster.Units.Macros
{
	using CT = CompileTimeUtil;

	/// Declares unit types with convertion rules
	macro unittypes(namespaceName, body)
	syntax ("unittypes", namespaceName, body)
	{
		// Parse namespace name
		def namespaceName = match (namespaceName)
		{
			| <[ $(n : name) ]> => n.Id
			| _ => Message.FatalError($"Invalid unittypes syntax: expected namespace name part, got $namespaceName")
		};
		
		// Prepare normalization coefficients map
		def coefSubMap = Dictionary();
		
		// Parse units
		match (body)
		{
			| <[ { .. $unitDecls } ]> =>
				// Check if BaseUnits must be initialized
				def initBaseUnits = match (CT.BaseUnits)
				{
					| null =>
						CT.BaseUnits = [];
						true
					| _ => false
				};

				mutable units = [];
				
				// Process all phys types and add info into PhysTypeHelper
				mutable index = 0;
				mutable calcBaseUnitPowers = initBaseUnits;
				foreach (<[ $decl ]> in unitDecls)
				{
					def unitCoef = match (decl)
					{
						| <[ $(n : name) ]> => (n.Id, 1.0)
						| <[ $(n : name) : ($coefExpr) ]> =>
							// Function parses simple coefficient exprs
							def parseCoefExpr(expr)
							{
								| <[ $(d : double) ]> => d
								| <[ $lhs * $(d : double) ]> => parseCoefExpr(lhs) * d
								| <[ $lhs / $(d : double) ]> => parseCoefExpr(lhs) / d
								| _ => Message.FatalError($"Invalid unittypes syntax: expected \"UnitName [: (coefficient expr)]\", got $decl")
							};
							(n.Id, parseCoefExpr(coefExpr))
						| <[ $(n : name) = $aliasExpr ]> =>
							// Aliases are started - remember it
							when (calcBaseUnitPowers)
							{
								// Fill base units list
								CT.BaseUnits = units.Reverse();

								// Prepare powers lists
								CT.BaseUnits.Iter(fun(unitName) { CT.PowersMap[unitName] = CT.ModifyBaseUnitPower([], unitName, _ + 1) });

								calcBaseUnitPowers = false
							};

							// Prepare powers list
							def oldN = n.Id == "SpecificGravity";
							def parseAliasExpr(expr)
							{
								| <[ $(n : name) ^ $(x : int) ]> => List.Map(CT.PowersMap[n.Id], _ * x)
								| <[ $lhs * $(n : name) ^ $(x : int) ]> =>
									List.Map2(parseAliasExpr(lhs), parseAliasExpr(<[ $(n : name) ^ $(x : int) ]>), _ + _)
								| <[ $lhs / $(n : name) ^ $(x : int) ]> =>
									List.Map2(parseAliasExpr(lhs), parseAliasExpr(<[ $(n : name) ^ $(x : int) ]>), _ - _)
								| <[ $(n : name) ]> => parseAliasExpr(<[ $(n : name) ^ 1 ]>)
								| <[ $lhs * $(n : name) ]> => parseAliasExpr(<[ $lhs * $(n : name) ^ 1 ]>)
								| <[ $lhs / $(n : name) ]> => parseAliasExpr(<[ $lhs / $(n : name) ^ 1 ]>)
								| _ => Message.FatalError($"Invalid unittypes syntax: expected \"UnitName = alias\", got $decl")
							};
							CT.PowersMap[n.Id] = parseAliasExpr(aliasExpr);

							(n.Id, 1.0)
						| _ => Message.FatalError($"Invalid unittypes syntax: expected \"UnitName [: (coefficient expr)]\", got $decl")
					};

					units ::= unitCoef[0];

					// Add new normalization coefficient into map
					coefSubMap[unitCoef[0]] = unitCoef[1];

					index++
				};

				units = units.Reverse();
				when (initBaseUnits)
				{
					CT.AllUnits = units
				}

				// Generate corresponding concrete types
				def units = if (units.Length >= CT.BaseUnits.Length) CT.AllUnits else units;
				foreach (n in units)
				{
					def powers = CT.PowersMap[n];

					mutable coef = 0.0;
					when (!coefSubMap.TryGetValue(n, out coef))
					{
						coef = 1.0;

						// Calculate coefficient from base unit coefficients on the fly
						List.Iter2(
							powers,
							CT.BaseUnits,
							fun (power : int, name)
							{
								when (power != 0)
								{
									def f = if (power > 0) _ * _ else _ / _;
									coef = f(coef, Math.Pow(coefSubMap[name], Math.Abs(power)));
								}
							})
					}

					_ = CT.GetUnitType(
						GlobalEnv.Core,
						$"$(CT.UnitsNamespace).$namespaceName.$n",
						powers,
						coef)
				}
			| _ => Message.FatalError($"Invalid unittypes syntax: expected semicolon-separated type names, got $body")
		};
		
		<[ () ]>
	}
}
