using System.Console;
using Oyster.Units;
using Oyster.Units.Macros;

// SI
unittypes Si {
	// Base units
	Mass;
	Length;
	Time;
	Temperature;
	CurrentStrength;
	LightIntensity;
	QuantityOfSubstance;

	// Aliases
	Area = Length ^ 2;
	Volume = Area * Length;
	Velocity = Length / Time;
	Acceleration = Velocity / Time;
	AngularVelocity = Velocity;
	AngularAcceleration = Acceleration;
	Density = Mass / Volume;

	Force = Mass * Acceleration;
	Pressure = Mass / Length / Time ^ 2;
	SpecificGravity = Pressure / Length;

	LinearMomentum = Mass * Velocity;
	MomentOfInertia = Mass * Area;

	Energy = Force * Length;
	Power = Energy / Time;

	DynamicViscosity = Mass / Length / Time;
	KinematicViscosity = DynamicViscosity;

	HeatCapacity = Energy / Temperature;
	Entropy = HeatCapacity;
	SpecificHeat = HeatCapacity / Mass;
	SpecificEntropy = SpecificHeat;

	Charge = Time * CurrentStrength;
	LinearDensityOfCharge = Charge / Length;
	SurfaceDensityOfCharge = Charge / Area;
	ElectricFluxDensity = SurfaceDensityOfCharge;
	PackedDensityOfCharge = Charge / Volume;

	Voltage = Power / CurrentStrength;
	ElectricForce = Force / CurrentStrength;
	Resistance = Voltage / CurrentStrength;
	SpecificResistance = Resistance * Length;

	Permittance = Time / Resistance;

	CurrentDensity = CurrentStrength / Length ^ 2
};

// CGS
unittypes Cgs {
	Mass : (1.0 / 1000.0);
	Length : (1.0 / 100.0);
	Time;
	Temperature;
	CurrentStrength : (1.0 / 3000000000.0);
	LightIntensity;
	QuantityOfSubstance
};

// EXT
unittypes Ext {
	Power : (1471.0 / 2.0)
};

// *****************************************************************************

def m1 = Cgs.Mass(1); // 1 
def m2 = Si.Mass(m1); //    .    0.001 

WriteLine($"Mass in SI: $m2, in CGS: $m1");

def l1 = Cgs.Length(100); // 100 
def l2 = Si.Length(l1);   //  .  - 1 

WriteLine($"Length in SI: $l2, in CGS: $l1");

def t1 = Cgs.Time(1); // 1 
def t2 = Si.Time(t1); //  1 , ..         .

WriteLine($"Time in SI: $t2, in CGS: $t1");

//     .   (, , ) 
//  ()    .         -
// ,   
def v1 = Cgs.Velocity(10); // 10 /
def v2 = Si.Velocity(v1);  //   - 0.1 /

WriteLine($"Velocity in SI: $v2, in CGS: $v1");

def e1 = Si.Energy(0.0001); // 10e-4 .
def e2 = Cgs.Energy(e1);    // e2 = 1000 

WriteLine($"Energy in SI: $e1, in CGS: $e2");

//  -     .       
//   
def p1 = Si.Power(1000); // 1000 
def p2 = Ext.Power(p1);  // 1.35962 

WriteLine($"Power in SI: $p1, in Ext system: $p2");

//      
def a1 = Si.Acceleration(1); // 1 /2
def a2 = Cgs.Acceleration(a1 + Cgs.Acceleration(200)); // 1 /2 + 200 /2 = 300 /2,     

WriteLine($"Acceleration: a1 = $a1, a2 = $a2");

// 
def f1 = Si.Force(10);      // 10 
def f2 = Cgs.Force(f1*2);  // 10  * 2 = 2000000 

def f3 = Si.Force(m1*l2/(t1*t1)); // 1  * 1  / (1  * 1 ) = 0.001 
//       -   
//def f4 = Si.Force(m1*l2/t1);
//    ncc:
// Main.n(122,10,122,18): error : each overload has an error during call:
// Main.n(122,10,122,18): error : overload #1, method Oyster.Units.Si.Force..ctor(value : System.Double) : void
// Main.n(122,10,122,18): error : in argument #1 (value) of Oyster.Units.Si.Force, needed a System.Double, got _N_TempUnitType5439: _N_TempUnitType5439 is not a subtype of System.Double [simple require]
// Main.n(122,10,122,18): error : overload #2, method Oyster.Units.Si.Force..ctor(value : Oyster.Units.AbstractUnitType_1_1_-2_0_0_0_0) : void
// Main.n(122,10,122,18): error : in argument #1 (value) of Oyster.Units.Si.Force, needed a Oyster.Units.AbstractUnitType_1_1_-2_0_0_0_0, got _N_TempUnitType5439: _N_TempUnitType5439 is not a subtype of Oyster.Units.AbstractUnitType_1_1_-2_0_0_0_0 [simple require]

WriteLine($"Force: f1 = $f1, f2 = $f2, f3 = $f3");

//    -   ,   
//WriteLine($"ERROR = $(e1 + e2 + v1 + v2 + m1 + m2)");
//    ncc (,  ):
// Main.n(135,52,135,64): error : Expression { e1 + e2 [+] v1 } can't be compiled because units are having different dimensions.

//     .    0.0002  -     , 
//    
WriteLine($"Result = $(e1 + e2)");

//   , .. e1  e2      ,     
//  
WriteLine($"Result = $(e1/e2)")
